import json
from collections import defaultdict
from dataclasses import dataclass
from os import environ

from gqlclient import GraphQLClient

from dkist_processing_common._util.config import get_mesh_config


@dataclass
class RecipeRunMutation:
    recipeRunId: int
    recipeRunStatusId: int
    authToken: str = environ.get("GQL_AUTH_TOKEN")


@dataclass
class RecipeRunStatusResponse:
    recipeRunStatusId: int


@dataclass
class RecipeRunStatusQuery:
    recipeRunStatusName: str


@dataclass
class CreateRecipeRunStatusResponse:
    recipeRunStatus: RecipeRunStatusResponse


@dataclass
class RecipeRunStatusMutation:
    recipeRunStatusName: str
    isComplete: bool
    recipeRunStatusDescription: str = "To add a description, use the processing-preparation-worker"
    authToken: str = environ.get("GQL_AUTH_TOKEN")


@dataclass
class InputDatasetResponse:
    inputDatasetDocument: str


@dataclass
class ProcessingCandidateResponse:
    observingProgramExecutionId: str
    proposalId: str


@dataclass
class RecipeInstanceResponse:
    inputDataset: InputDatasetResponse
    processingCandidate: ProcessingCandidateResponse


@dataclass
class RecipeRunResponse:
    recipeInstance: RecipeInstanceResponse


@dataclass
class RecipeRunInputDatasetQuery:
    recipeRunId: int


@dataclass
class DatasetCatalogReceiptAccountMutation:
    """
    Dataclass used to write the dataset_catalog_receipt_account record for the run.
    It sets an expected object count for a dataset so that dataset inventory creation
    doesn't happen until all objects are transferred and inventoried.
    """

    datasetId: str
    expectedObjectCount: int
    authToken: str = environ.get("GQL_AUTH_TOKEN")


recipe_run_statuses = {
    "INPROGRESS": "Recipe run is currently undergoing processing",
    "COMPLETEDSUCCESSFULLY": "Recipe run processing completed with no errors",
}


# Environment variable indicating how to connect to dependencies on the service mesh
MESH_CONFIG = get_mesh_config()

graph_ql_client = GraphQLClient(
    f'http://{MESH_CONFIG["internal-api-gateway"]["mesh_address"]}:{MESH_CONFIG["internal-api-gateway"]["mesh_port"]}/graphql'
)


def get_message_status_query(status: str):
    """
    Find the id of a recipe run status
    """
    return graph_ql_client.execute_gql_query(
        query_base="recipeRunStatuses",
        query_response_cls=RecipeRunStatusResponse,
        query_parameters=RecipeRunStatusQuery(recipeRunStatusName=status),
    )


def add_new_recipe_run_status(status: str, is_complete: bool) -> int:
    """
    Add a new recipe run status to the db
    :param status: name of the status to add
    :param is_complete: does the new status correspond to an accepted completion state
    """
    if not isinstance(status, str):
        raise TypeError(f"status must be of type str: {status}")
    if not isinstance(is_complete, bool):
        raise TypeError(f"is_complete must be of type bool: {is_complete}")
    recipe_run_status_response = graph_ql_client.execute_gql_mutation(
        mutation_base="createRecipeRunStatus",
        mutation_response_cls=CreateRecipeRunStatusResponse,
        mutation_parameters=RecipeRunStatusMutation(
            recipeRunStatusName=status,
            isComplete=is_complete,
            recipeRunStatusDescription=recipe_run_statuses[status],
        ),
    )
    return recipe_run_status_response.recipeRunStatus.recipeRunStatusId


def apply_status_id_to_recipe_run(recipe_run_status_id: int, recipe_run_id: int):
    """
    Change the status of a given recipe run id
    :param recipe_run_status_id: the new status to use
    :param recipe_run_id: id of the recipe run to have the status changed
    """
    graph_ql_client.execute_gql_mutation(
        mutation_base="updateRecipeRun",
        mutation_parameters=RecipeRunMutation(
            recipeRunId=recipe_run_id, recipeRunStatusId=recipe_run_status_id
        ),
    )



import logging
from os import environ
from pathlib import Path

from astropy.io import fits

from dkist_processing_common._util.globus import submit_globus_transfer
from dkist_processing_common._util.graphql import DatasetCatalogReceiptAccountMutation
from dkist_processing_common._util.graphql import graph_ql_client
from dkist_processing_common._util.message import publish_messages
from dkist_processing_common.base import SupportTaskBase


class TransferInputData(SupportTaskBase):
    """
    Changes the status of the recipe run to "INPROGRESS"
    Starts a globus transfer of all data listed in the associated input dataset to scratch
    """

    def run(self) -> None:
        self.change_status_to_in_progress()

        bucket = self.input_dataset(section="bucket")
        frames = self.input_dataset(section="frames")

        source_files = [Path("/", bucket, frame) for frame in frames]

        submit_globus_transfer(
            source_files=source_files,
            destination_files=[
                self.globus_path(self.input_dir, source_file.name) for source_file in source_files
            ],
            source_endpoint=environ.get("OBJECT_STORE_ENDPOINT"),
            destination_endpoint=environ.get("SCRATCH_ENDPOINT"),
        )


class TagInputData(SupportTaskBase):
    """
    Tags all input data with "INPUT" and whatever is listed in "DKIST004" in the headers
    """

    def run(self) -> None:
        input_file_paths = self.input_dir.glob("*.fits")
        for f in input_file_paths:
            header = fits.open(str(f))[0].header
            self.tag(str(f.resolve()), "INPUT")
            self.tag(str(f.resolve()), header["DKIST004"].upper())


class TransferOutputData(SupportTaskBase):
    """
    Transfers all data tagged as "OUTPUT" to the object store
    """

    def run(self) -> None:
        globus_output_paths = [self.globus_path(path) for path in self.output_paths]
        logging.info(f"{globus_output_paths=}")
        destination_files = [
            Path("/data", self.proposal_id, self.dataset_id, source_file.name)
            for source_file in globus_output_paths
        ]
        logging.info(f"{destination_files=}")

        submit_globus_transfer(
            source_files=globus_output_paths,
            destination_files=destination_files,
            source_endpoint=environ["SCRATCH_ENDPOINT"],
            destination_endpoint=environ["OBJECT_STORE_ENDPOINT"],
        )


class AddDatasetReceiptAccount(SupportTaskBase):
    """
    Creates a Dataset Catalog Receipt Account record populated with the number of objects
    """

    def run(self) -> None:
        expected_object_count = len(self.output_paths)
        graph_ql_client.execute_gql_mutation(
            mutation_base="createDatasetCatalogReceiptAccount",
            mutation_parameters=DatasetCatalogReceiptAccountMutation(
                datasetId=self.dataset_id, expectedObjectCount=expected_object_count
            ),
        )


class PublishCatalogMessages(SupportTaskBase):
    """
    Publishes a message for every object transferred to the object store
    """

    def run(self) -> None:
        object_filepaths = [
            Path(self.proposal_id, self.dataset_id, source_file.name)
            for source_file in self.output_paths
        ]
        messages = []
        for filepath in object_filepaths:
            if filepath.lower().endswith(".fits"):
                messages.append(self.create_frame_message(object_filepath=filepath))
            elif filepath.lower().endswith(".mp4"):
                messages.append(self.create_movie_message(object_filepath=filepath))
        publish_messages(messages=messages)


class Teardown(SupportTaskBase):
    """
    Changes the status of the recipe run to "COMPLETEDSUCCESSFULLY"
    Deletes the scratch directory containing all data from this pipeline run
    """

    def run(self) -> None:
        logging.info(f"Removing data and tags for recipe run {self.recipe_run_id}")
        self.change_status_to_completed_successfully()
        self.purge_file_system()



"""
Tag cloud manager
"""
from typing import Iterable
from typing import List
from typing import Set
from typing import Tuple
from typing import Union

from redis import Redis

from dkist_processing_common._util.config import get_mesh_config


class TagDB:
    def __init__(self, recipe_run_id: int, task_name: str):
        """
        Initialize a connection to a tag database whose instance is uniquely identified by
        the recipe_run_id and whose client connection is identified with the task_name
        """
        self.recipe_run_id = recipe_run_id
        self.task_name = task_name
        host, port = self._db_host_port()
        self._db = self.recipe_run_id % 15
        self._name_space = f"{self.recipe_run_id}:"
        self.db = Redis(db=self._db, host=host, port=port, client_name=self.task_name)

    @staticmethod
    def _db_host_port() -> Tuple[str, int]:
        mesh_default = '{"automated-processing-scratch-inventory": {"mesh_address": "localhost", "mesh_port": 6379}}'
        mesh_config = get_mesh_config(default=mesh_default)
        return (
            mesh_config["automated-processing-scratch-inventory"]["mesh_address"],
            mesh_config["automated-processing-scratch-inventory"]["mesh_port"],
        )

    @staticmethod
    def _format_query_result(result: List[bytes]) -> Set[str]:
        return {r.decode("utf8") for r in result}

    def _format_name_space(self, tags: Union[Iterable[str], str]) -> Union[Iterable[str], str]:
        if isinstance(tags, str):
            return self._name_space + tags
        return [f"{self._name_space}{t}" for t in tags]

    def add(self, tag: str, value: str):
        """
        Add values to a tag
        """
        tag = self._format_name_space(tag)
        self.db.sadd(tag, value)

    def any(self, tags: Union[Iterable[str], str]) -> Set[str]:
        """
        Return a set of values that match any of the tags
        """
        tags = self._format_name_space(tags)
        r = self.db.sunion(tags)
        return self._format_query_result(r)

    def all(self, tags: Union[Iterable[str], str]) -> Set[str]:
        """
        Return a set of values that match all of the tags
        """
        tags = self._format_name_space(tags)
        r = self.db.sinter(tags)
        return self._format_query_result(r)

    def close(self):
        """
        Close the connection to the tag db.  For use at the end of a task
        """
        self.db.close()

    @property
    def _namespace_keys(self) -> List[str]:
        return self.db.keys(f"{self._name_space}*")

    def purge(self) -> None:
        """
        Remove the database of tags.  For use at the end of a workflow.
        """
        if keys := self._namespace_keys:
            self.db.delete(*keys)

    def __repr__(self):
        return f"TagDB(recipe_run_id={self.recipe_run_id}, task_name={self.task_name})"

    def __str__(self):
        return f"{self!r} connected to {self.db}"



"""
Global test fixtures
"""
from pathlib import Path
from random import randint
from typing import Tuple

import pytest
from astropy.io import fits

from dkist_processing_common._util.scratch import WorkflowFileSystem
from dkist_processing_common._util.tags import TagDB
from dkist_processing_common.base import ScienceTaskL0ToL1
from dkist_processing_common.base import SupportTaskBase


@pytest.fixture(scope="function")
def support_task():
    """
    Create task class for usage in tests
    """

    class TaskClass(SupportTaskBase):
        def run(self) -> None:
            pass

    i = randint(0, 1000000000)

    task = TaskClass(
        recipe_run_id=i,
        workflow_name=f"workflow_name_{i}",
        workflow_version=f"version{i}",
    )

    yield task

    task._scratch.purge()


@pytest.fixture(scope="function")
def science_task():
    """
    Create task class for usage in tests
    """

    class TaskClass(ScienceTaskL0ToL1):
        def run(self) -> None:
            pass

    i = randint(0, 1000000000)

    task = TaskClass(
        recipe_run_id=i,
        workflow_name=f"workflow_name_{i}",
        workflow_version=f"version{i}",
    )

    yield task

    task._scratch.purge()


@pytest.fixture(scope="function")
def science_task_with_inputs(science_task, tmp_path):
    datatypes = ["DARK", "GAIN", "TARGET", "INSTPOLCAL", "TELPOLCAL", "GEOMETRIC"]
    for datatype in datatypes:
        hdu = fits.PrimaryHDU()
        hdu.header["DKIST004"] = datatype
        hdul = fits.HDUList([hdu])
        science_task._scratch = WorkflowFileSystem(scratch_base_path=tmp_path)
        science_task._scratch.write_fits(hdul, f"{datatype}.fits", tags=["INPUT", datatype])
    return science_task


@pytest.fixture(scope="function")
def support_task_with_outputs(support_task, tmp_path):
    paths = [f"output/{i}.fits" for i in range(10)]
    for path in paths:
        hdu = fits.PrimaryHDU()
        hdul = fits.HDUList([hdu])
        support_task._scratch = WorkflowFileSystem(scratch_base_path=tmp_path)
        support_task._scratch.write_fits(hdul, path, tags=["OUTPUT"])
    return support_task


@pytest.fixture(scope="session")
def input_dark_fits():
    hdu = fits.PrimaryHDU()
    hdu.header["DKIST004"] = "DARK"
    return fits.HDUList([hdu])


@pytest.fixture(scope="session")
def intermediate_fits():
    hdu = fits.PrimaryHDU()
    hdu.header["DKIST004"] = "TEST_INTERMEDIATE"
    return fits.HDUList([hdu])


@pytest.fixture(scope="session")
def output_fits():
    hdu = fits.PrimaryHDU()
    hdu.header["DKIST004"] = "OBSERVE"
    return fits.HDUList([hdu])


@pytest.fixture()
def recipe_run_id():
    return randint(0, 99999)


@pytest.fixture()
def tag_db(recipe_run_id) -> TagDB:
    t = TagDB(recipe_run_id=recipe_run_id, task_name="test_tags")
    yield t
    t.purge()
    t.close()


@pytest.fixture()
def tag_db2(recipe_run_id) -> TagDB:
    """
    Another instance of a tag db in the same redis db
    """
    recipe_run_id = recipe_run_id + 15  # same db number but different namespace
    t = TagDB(recipe_run_id=recipe_run_id, task_name="test_tags2")
    yield t
    t.purge()
    t.close()


@pytest.fixture(params=[None, "use_tmp_path"])
def workflow_file_system(request, recipe_run_id, tmp_path) -> Tuple[WorkflowFileSystem, int, Path]:
    if request.param == "use_tmp_path":
        path = tmp_path
    else:
        path = request.param
    wkflow_fs = WorkflowFileSystem(
        recipe_run_id=recipe_run_id,
        task_name="wkflow_fs_test",
        scratch_base_path=path,
    )
    yield wkflow_fs, recipe_run_id, tmp_path
    wkflow_fs.purge()
    tmp_path.rmdir()
    wkflow_fs.close()



import json
from collections import defaultdict
from os import environ
from uuid import uuid4

from talus.message import message_class

from dkist_processing_common._util.config import get_mesh_config

# Environment variable indicating how to connect to dependencies on the service mesh
MESH_CONFIG = get_mesh_config()

RABBITMQ_CONFIG = {
    "rabbitmq_host": MESH_CONFIG["interservice-bus"]["mesh_address"],
    "rabbitmq_port": MESH_CONFIG["interservice-bus"]["mesh_port"],
    "rabbitmq_user": environ.get("ISB_USERNAME"),
    "rabbitmq_pass": environ.get("ISB_PASSWORD"),
}


@message_class(routing_key="catalog.frame.m", queues=["catalog.frame.q"])
class CatalogFrameMessage:
    """
    Class to hold the catalog_frame_message to be sent to RabbitMQ
    """

    objectName: str = "default_object_name"
    conversationId: str = str(uuid4().hex)
    bucket: str = "data"
    incrementDatasetCatalogReceiptCount: bool = True


@message_class(routing_key="catalog.object.m", queues=["catalog.object.q"])
class CatalogObjectMessage:
    """
    Class to hold the catalog_object_message to be sent to RabbitMQ
    """

    objectType: str
    objectName: str = "default_object_name"
    conversationId: str = str(uuid4().hex)
    bucket: str = "data"
    groupId: str = "default_group_id"
    groupName: str = "DATASET"
    incrementDatasetCatalogReceiptCount: bool = True
